import logging
import re

from remotemanager.storage.sendablemixin import SendableMixin

logger = logging.getLogger(__name__)


class Flags(SendableMixin):
    """
    Basic but flexible handler for terminal command flags

    Allows for inplace modification:

    >>> f = Flags('abcdd')
    >>> f -= 'd'
    >>> f.flags
    >>> '-abcd'

    Arguments:
        initial_flags (str):
            initial base flags to be used and modified if needed
    """

    def __init__(self, *initial_flags):
        if not isinstance(initial_flags, str):
            initial_flags = " ".join(initial_flags)

        logger.debug("creating Flags with initial flags %s", initial_flags)

        self._flags = {}
        self.flags = initial_flags

    def __repr__(self):
        return self.flags

    def _add(self, section):
        raw, prefix = self.parse_string(section)
        if prefix == "-":
            self.ensure_prefix_exists(prefix)
            self._flags["-"] += list(raw)
        else:
            self.ensure_prefix_exists(prefix)
            self._flags[prefix].append(raw)

        logger.debug("adding %s to flags. Flags are now %s", raw, self.flags)

    def __add__(self, other):
        sections = other.split(" ")
        for section in sections:
            self._add(section)

    def __iadd__(self, other):
        self.__add__(other)
        logger.debug("adding %s to flags inplace. Flags are now %s", other, self.flags)
        return self

    def _sub(self, section):
        raw, prefix = self.parse_string(section)
        if prefix == "-":
            for char in raw:
                try:
                    self._flags["-"].remove(char)
                except ValueError:
                    pass
        else:
            try:
                self._flags["--"].remove(raw)
            except ValueError:
                pass

        logger.debug("subtracting %s from flags. Flags are now %s", raw, self.flags)

    def __sub__(self, other):
        """Subtract unique flags in `other` once."""
        sections = other.split(" ")
        for section in sections:
            self._sub(section)

    def __isub__(self, other):
        self.__sub__(other)
        logger.debug(
            "subtracting %s from flags inplace. Flags are now %s", other, self.flags
        )
        return self

    def ensure_prefix_exists(self, prefix):
        """Ensures that the prefix exists in the internal storage, creating
        it if not"""
        if prefix not in self._flags:
            self._flags[prefix] = []

    def parse_string(self, string) -> [str, bool]:
        """
        Takes a string, and strips away any non-alphanumeric chars.
        Returns True in secondary return if this is a verbose flag

        Args:
            string (str):
                input flags

        Returns (str, bool):
            filtered input
            True if this is a verbose flag
        """
        raw = strip_non_alphanumeric(string)
        num = string.count("-")

        # special case
        if num == 0:
            num = 1

        return raw, "-" * num

    @property
    def flags(self):
        """Returns the fully qualified flags as a string"""
        if sum([len(f) for f in self._flags.values()]) == 0:
            logger.info("sum of lens is 0, returning ''")
            return ""
        logger.info("creating string from internal flags %s", self._flags)
        output = []
        for prefix, flags in self._flags.items():
            if prefix == "-":
                output.append("-" + "".join(flags))
            else:
                output.append(prefix + f" {prefix}".join(flags))

        string = " ".join(output)
        logger.info("done -> %s", string)
        return string

    @flags.setter
    def flags(self, inp):
        """Set the flags to the new input

        Arguments:
            inp (str):
                new flags to use (will overwrite old ones)
        """
        self._flags = {}
        self.__add__(inp)


def strip_non_alphanumeric(string):
    """
    remove any non-alphanumeric strings from input string

    Args:
        string (str):
            input string

    Returns (str):
        input string, sans any non-alphanumeric chars

    """
    pattern = re.compile(r"[\W_]+", re.UNICODE)  # noqa: W605

    return pattern.sub("", string)



import logging
import os.path
from typing import Union

from remotemanager.connection.cmd import CMD
from remotemanager.connection.computers.base import BaseComputer
from remotemanager.storage.trackedfile import TrackedFile

logger = logging.getLogger(__name__)


class Script(BaseComputer):
    """
    Subclass of BaseComputer providing the ability to directly execute the scripts
    """

    def __init__(
        self,
        template: str,
        local_dir: str = ".",
        remote_dir: str = ".",
        filename: Union[None, str] = None,
        **kwargs,
    ):
        super().__init__(template, **kwargs)
        filename = filename or f"{abs(self.uuid)}.sh"

        self._file = TrackedFile(local_dir, remote_dir, filename)
        self._cmd = None

    def __hash__(self) -> int:
        return self.uuid

    @property
    def uuid(self) -> int:
        """Returns the hash of the template"""
        return hash(self.template)

    @property
    def exec_cmd(self) -> CMD:
        """Returns the CMD instance that was used to execute the script"""
        return self._cmd

    @property
    def local_dir(self) -> str:
        """Returns the local dir"""
        return self._file.local_dir

    @local_dir.setter
    def local_dir(self, path: str):
        """Sets the local dir, if possible"""
        if self.exec_cmd is not None:
            raise ValueError("Cannot set path on an executed run, set at run()")
        self._file._local_path = path

    @property
    def remote_dir(self) -> str:
        """Returns the remote dir"""
        return self._file.remote_dir

    @remote_dir.setter
    def remote_dir(self, path: str):
        """Sets the remote dir, if possible"""
        if self.exec_cmd is not None:
            raise ValueError("Cannot set path on an executed run, set at run()")
        self._file._remote_path = path

    @property
    def local_path(self) -> str:
        """Full local path"""
        return self._file.local

    @property
    def remote_path(self) -> str:
        """Full remote path"""
        return self._file.remote

    @property
    def filename(self) -> str:
        """Returns the filename"""
        return self._file.name

    @property
    def file(self) -> TrackedFile:
        """Returns the TrackedFile instance for this script"""
        return self._file

    def run(
        self,
        exectuable: str = "bash",
        local_dir: Union[None, str] = None,
        remote_dir: Union[None, str] = None,
        stream: bool = False,
        cleanup: bool = True,
        **kwargs,
    ) -> str:
        """
        Attempt to run the current script on the remote

        Args:
            exectuable (str): exectuable to call
            local_dir (str): update local_dir
            remote_dir (str): update remote_dir
            stream (bool): stream output
            cleanup (bool): Delete files after execution if True
            kwargs (dict): args to pass to script() method
        """
        self._cmd = None
        if local_dir is not None:
            self.local_dir = local_dir
        if remote_dir is not None:
            self.remote_dir = remote_dir

        script = self.script(**kwargs)

        self.file.write(script)

        if self.local_path != self.remote_path:
            self.transport.queue_for_push(self.file)
            self.transport.transfer()
        else:
            logger.debug("local and remote paths are equal, ignoring transfer")

        cmd = [f"cd {self.file.remote_dir} && {exectuable} {self.file.name}"]

        if cleanup:
            cmd.append(f" && rm {self.file.remote}")

        self._cmd = self.cmd(
            "".join(cmd),
            stream=stream,
            raise_errors=False,
        )
        self._cmd.exec()

        if cleanup:
            try:
                os.remove(self.file.local)
            except FileNotFoundError:
                pass

        return self.exec_cmd.stdout

    @property
    def stdout(self) -> Union[None, str]:
        """Returns the stdout of the run, if available"""
        if self.exec_cmd is not None:
            return self.exec_cmd.stdout
        return None

    @property
    def stderr(self) -> Union[None, str]:
        """Returns the stderr of the run, if available"""
        if self.exec_cmd is not None:
            return self.exec_cmd.stderr
        return None



"""Verbosity enables deeper verbose levels than True and False"""

from typing import Union

from remotemanager.storage.sendablemixin import SendableMixin


class Verbosity(SendableMixin):
    """
    Class to store verbosity information

    Initialise with Verbosity(level), where level is the integer level

    Printing can be requested with Verbose.print(msg, level)

    If the verbose level is set above `level`, the message will be printed

    args:
        level (int, bool, Verbosity):
            level above which to print
    """

    _default_level = 1
    __slots__ = ["level"]

    def __init__(self, level: Union[None, int, bool, "Verbosity"] = None):
        if level is None:
            level = Verbosity._default_level

        # see if the level passed is already a Verbose instance
        if isinstance(level, self.__class__):
            level = level.value
        elif isinstance(level, bool):
            level = int(level)

        self.level = level

    def __repr__(self) -> str:
        return f"Verbosity({self.level})"

    def __bool__(self) -> bool:
        return self.level != 0

    def _prepare_other_for_comparison(self, other: Union[int, "Verbosity"]) -> int:
        if isinstance(other, self.__class__):
            return other.level
        return other

    def __eq__(self, other: Union[int, "Verbosity"]) -> bool:
        return self.level == self._prepare_other_for_comparison(other)

    def __lt__(self, other: Union[int, "Verbosity"]) -> bool:
        return self.level < self._prepare_other_for_comparison(other)

    def __le__(self, other: Union[int, "Verbosity"]) -> bool:
        return self.level <= self._prepare_other_for_comparison(other)

    def __gt__(self, other: Union[int, "Verbosity"]) -> bool:
        return self.level > self._prepare_other_for_comparison(other)

    def __ge__(self, other: Union[int, "Verbosity"]) -> bool:
        return self.level >= self._prepare_other_for_comparison(other)

    @property
    def value(self) -> int:
        """Alias for self.level"""
        return self.level

    def print(self, message: str, level: int, end: str = "\n"):
        """
        Request that a message be printed. Compares against the set
        verbosity level before printing.

        Args:
            message (str):
                message to print
            level (int):
                If this number is higher priority than (lower numeric value)
                (or equal to) the set limit, print
            end (str):
                print(..., end= ...) hook
        """
        # print(f'request {message[:24]} @ {atlevel}')
        if self and self.level >= level:
            print(message, end=end)



import logging
import warnings
from typing import Union, Any

from remotemanager.connection.computers.dynamicvalue import DynamicMixin

logger = logging.getLogger(__name__)


class Substitution(DynamicMixin):
    """
    Stores a jobscript template substitution

    Args:
        target:
            String to _replace_ in the template.
            Ideally should begin with a commenting character (#)
        name:
            String to replace _with_. At script generation, `target`
            will be replaced with `name`

    .. note::
        For other args see the DynamicMixin class
    """

    __slots__ = ["target", "mode", "executed", "dependencies"]

    def __init__(
        self,
        target: str,
        name: str,
        default: Union[Any, None] = None,
        value: Union[Any, None] = None,
        optional: bool = True,
        requires: Union[str, list, None] = None,
        replaces: Union[str, list, None] = None,
        min: Union[int, None] = None,
        max: Union[int, None] = None,
        mode: Union[str, None] = None,
        format: Union[str, None] = None,
    ):
        super().__init__(
            assignment=name,
            default=default,
            value=value,
            optional=optional,
            requires=requires,
            replaces=replaces,
            min=min,
            max=max,
            format=format,
        )

        self.target = target
        self.mode = mode

        self.executed = False
        self.dependencies = []  # must calculate the value of these first

    def __hash__(self) -> int:
        return hash(self.target + self.name)

    def __str__(self):
        return str(self.value)

    def __repr__(self) -> str:
        return f"Substitution({self.target}, {self.name}) -> {self.value}"

    def __bool__(self):
        """
        Makes objects "falsy" if no value has been set, "truthy" otherwise
        """
        return self.value is not None

    @classmethod
    def from_string(
        cls, string: str, warn_invalid: bool = True, **kwargs
    ) -> "Substitution":
        """
        Create a substitution object from template string

        Args:
            string (str): Input string to generate from
            warn_invalid (bool): Invalid args name, target will be deleted. Warn if True
            kwargs: Any keyword args to override the string with
        """

        content = string.strip("#")  # actual internal "content"
        sym, *args = content.split(":")  # name, then any : separated args

        logger.debug("\ttarget=%s, symbol=%s, args=%s", string, sym, args)

        string_args = {}
        for arg in args:
            logger.debug("\t\tprocessing arg %s", arg)
            if arg == "":
                logger.debug("\t\t\targ is empty")
                continue
            key, val = arg.split("=")

            string_args[key] = val

        invalid_args = ["name", "target"]
        for arg in invalid_args:
            if arg in kwargs:
                if warn_invalid:
                    warnings.warn(f"Invalid kwarg {arg}={kwargs[arg]} deleted")
                del kwargs[arg]

        string_args.update(kwargs)

        name = sym.lower()

        logger.debug("processing symbol: %s", string)

        return cls(name=name, target=string, **string_args)

    @property
    def target_kwargs(self):
        """Attempts to extract the kwargs from the target string"""
        args = self.target.strip("#").split(":")

        return {arg.split("=")[0]: arg.split("=")[1] for arg in args[1:]}

    @property
    def value(self) -> any:
        """
        Returns:
            value if present, else default
        """
        val = super().value

        if len(self.dependencies) == 0:
            return val
        # if we have dependents, we need to evaluate them
        evaluate = {}
        # int values cause this replace to fail
        # non strings probably shouldn't end up here,
        # but it's best not to error weirdly
        val = str(val)
        for dep in self.dependencies:
            evaluate[dep.arg] = dep.value
            # replace the target $value with the evaluated arg
            # print(f"replacing {dep.arg} -> {dep.value}")
            val = val.replace(f"${dep.arg}", str(dep.value))
        # now everything is replaced, we can freely eval
        # print(f"evaluating {val}")
        val = eval(str(val.strip()))
        # print(f"{self.arg}={val}\n")
        # cache the result
        self.value = val
        self.dependencies = {}

        return val

    @value.setter
    def value(self, value):
        super().set_value(value)

    @property
    def arg(self) -> str:
        """
        Returns:
            Argument which is exposed to the underlying URL for setting
        """
        return self.name.strip("$")

    @property
    def entrypoint(self) -> str:
        """Returns the name/name for this sub"""
        return self.name

    def pack(self, collect_value: bool = True) -> dict:
        """Store this Substitution in dict form"""
        data = super().pack(collect_value=collect_value)

        if self.mode is not None:
            data["mode"] = self.mode
        if self.target is not None:
            data["target"] = self.target

        return data
